'use strict';

var commands_js = require('./commands.js');

function git(dir) {
  return new Git(dir);
}

// The class is merely a fluent command/query builder
class Git {
  constructor(dir) {
    if (dir) {
      this.workdir = dir;
      this.gitdir = `${dir}/.git`;
    }
    this.operateRemote = 'origin';
    this.operateDepth = 0;
    this.operateBranch = 'master';
  }
  workdir(dir) {
    this.workdir = dir;
    return this;
  }
  gitdir(dir) {
    this.gitdir = dir;
    return this;
  }
  // This form works for basic username / password auth, or
  // the newer username / token auth that is often required if
  // 2FA is enabled.
  auth(username, password) {
    // Allow specifying it as one argument (mostly for CLI inputability)
    if (password === undefined) {
      let i = username.indexOf(':');
      if (i > -1) {
        password = username.slice(i + 1);
        username = username.slice(0, i);
      }
    }
    this.operateUsername = username;
    this.operatePassword = password || ''; // Enables the .auth(GITHUB_TOKEN) no-username shorthand
    return this;
  }
  // This for is for actual OAuth2 uses. Unfortunately, the
  // major players all have different conventions.
  oauth2(company, token) {
    switch (company) {
      case 'github':
        this.operateUsername = token;
        this.operatePassword = 'x-oauth-basic';
        break;
      case 'bitbucket':
        this.operateUsername = 'x-token-auth';
        this.operatePassword = token;
        break;
      case 'gitlab':
        this.operateUsername = 'oauth2';
        this.operatePassword = token;
        break;
      default:
        throw new Error(`I don't know how ${company} expects its Basic Auth headers to be formatted for OAuth2 usage. If you do, you can use the regular '.auth(username, password)' to set the basic auth header yourself.`);
    }
    return this;
  }
  remote(name) {
    this.operateRemote = name;
    return this;
  }
  author(name) {
    this.operateAuthorName = name;
    return this;
  }
  branch(name) {
    this.operateBranch = name;
    return this;
  }
  email(email) {
    this.operateAuthorEmail = email;
    return this;
  }
  datetime(date) {
    this.operateAuthorDateTime = date;
    return this;
  }
  depth(depth) {
    this.operateDepth = parseInt(depth);
    return this;
  }
  timestamp(seconds) {
    // seconds since unix epoch
    this.operateAuthorTimestamp = seconds;
    return this;
  }
  signingKey(asciiarmor) {
    this.privateKeys = asciiarmor;
    return this;
  }
  verificationKey(asciiarmor) {
    this.publicKeys = asciiarmor;
    return this;
  }
  outputStream(stream) {
    this.outputStream = stream;
    return this;
  }
  inputStream(stream) {
    this.inputStream = stream;
    return this;
  }
  async findRoot(dir) {
    return commands_js.findRoot(dir);
  }
  async init() {
    await commands_js.init(this.gitdir);
  }
  async fetch(ref) {
    await commands_js.fetch({
      gitdir: this.gitdir,
      ref,
      depth: this.operateDepth,
      remote: this.operateRemote,
      authUsername: this.operateUsername,
      authPassword: this.operatePassword
    });
  }
  async checkout(ref) {
    await commands_js.checkout({
      workdir: this.workdir,
      gitdir: this.gitdir,
      ref,
      remote: this.operateRemote
    });
  }
  async clone(url) {
    await commands_js.init(this.gitdir);
    // Add remote
    await commands_js.setConfig({
      gitdir: this.gitdir,
      path: 'remote.origin.url',
      value: url
    });
    // Fetch commits
    await commands_js.fetch({
      gitdir: this.gitdir,
      ref: `refs/heads/${this.operateBranch}`,
      depth: this.operateDepth,
      remote: this.operateRemote,
      authUsername: this.operateUsername,
      authPassword: this.operatePassword
    });
    // Checkout branch
    await commands_js.checkout({
      workdir: this.workdir,
      gitdir: this.gitdir,
      ref: this.operateBranch,
      remote: this.operateRemote
    });
  }
  async list() {
    return commands_js.list({
      gitdir: this.gitdir
    });
  }
  async add(filepath) {
    return commands_js.add({
      gitdir: this.gitdir,
      workdir: this.workdir,
      filepath
    });
  }
  async remove(filepath) {
    return commands_js.remove({
      gitdir: this.gitdir,
      filepath
    });
  }
  async commit(message) {
    return commands_js.commit({
      gitdir: this.gitdir,
      author: {
        name: this.operateAuthorName || (await this.getConfig('user.name')),
        email: this.operateAuthorEmail || (await this.getConfig('user.email')),
        timestamp: this.operateAuthorTimestamp,
        date: this.operateAuthorDateTime
      },
      committer: {
        name: this.operateAuthorName || (await this.getConfig('user.name')),
        email: this.operateAuthorEmail || (await this.getConfig('user.email')),
        timestamp: this.operateAuthorTimestamp,
        date: this.operateAuthorDateTime
      },
      message,
      privateKeys: this.privateKeys
    });
  }
  async verify(ref) {
    return commands_js.verify({
      gitdir: this.gitdir,
      publicKeys: this.publicKeys,
      ref
    });
  }
  async pack(oids) {
    return commands_js.pack({
      gitdir: this.gitdir,
      outputStream: this.outputStream,
      oids
    });
  }
  async unpack(oids) {
    return commands_js.unpack({
      gitdir: this.gitdir,
      inputStream: this.inputStream
    });
  }
  async push(ref) {
    let url = await commands_js.getConfig({
      gitdir: this.gitdir,
      path: `remote.${this.operateRemote}.url`
    });
    console.log('url =', url);
    return commands_js.push({
      gitdir: this.gitdir,
      ref,
      url,
      authUsername: this.operateUsername,
      authPassword: this.operatePassword
    });
  }
  async pull(ref) {
    return commands_js.fetch({
      gitdir: this.gitdir,
      ref,
      remote: this.operateRemote,
      authUsername: this.operateUsername,
      authPassword: this.operatePassword
    });
  }
  async getConfig(path) {
    return commands_js.getConfig({
      gitdir: this.gitdir,
      path
    });
  }
  async setConfig(path, value) {
    return commands_js.setConfig({
      gitdir: this.gitdir,
      path,
      value
    });
  }
  async status(pathname) {
    return commands_js.status({
      gitdir: this.gitdir,
      workdir: this.workdir,
      pathname
    });
  }
}

module.exports = git;
